%{
#include <config.h>

#include <stdio.h>
#include <string.h>

#include <map>

#include "line.h"
#include "scanner.h"
#include "discovery.h"
#include "util.h"
#include "global.h"

using namespace std;

static msn last_chat;

static char servername[40];

msn cfg_country;
msn cfg_area;
msn cfg_msn;

typedef map<const string, connection *, string_less_than> con_map;
con_map connections;
struct mask_map mask_map;
ctrl_connection control_connection;

void con_apply (const char *id, connection::message f, const char *msg = 0)
{
  con_map::iterator p = connections.find (id);
  if (p != connections.end ())
    {
      if (msg) printf ("%s -> %s: ", id, msg), fflush (stdout);
      (p->second->*f) ();
      if (msg) printf ("done\n");
    }
  else
    fprintf (stderr, "%s: no such connection\n", id);
}

#define CFG_TIMEOUT 5000

int num_connections ()
{
  int count = 0;
  con_map::const_iterator i = connections.begin ();
  
  while (i != connections.end ())
    count++, i++;
  
  return count;
}

/* wire functions */
packet ctrl_connection::send_packet ()
{
  packet p;
  
  lock.lock ();
  if (once ())
    {
      p = once;
      once.reset ();
      tx_at (5000);
    }
  else if (sent ())
    {
      if (ie_mask m = mask_map.dst_mask (sent.dst_msn))
        {
          sent.mask = m;
          p = sent;
          sent.reset ();
        }
    }
  
  lock.unlock ();
  return p;
}

void ctrl_connection::send_ctrl_poll (const char *src, const char *dst, u8 reply)
{
  lock.lock ();
  sent.set (src, dst, IEM_NONE);
  sent.put (MSG_CFG);
  sent.put (CFG_POLL | reply);
  sent.put (PROTOCOL_VERSION);
  
  sent.put (num_connections ());
  
  for(char *c = servername; *c; c++)
    sent.put (*c);
  
  sent.put (0);
  lock.unlock ();
}

void ctrl_connection::receive_packet (const packet &p)
{
  lock.lock ();
  
  disc_timeout.reset ();
  log (L_JUNK, "config packet %02x received\n", (int)p[0]);
  if (p[0] == MSG_DISCOVERY)
    {
      disc_timeout.reset ();
      
      mask_map.lock.lock ();
      ie_masks &m = mask_map[p.src_msn];
      mask_map.lock.unlock ();
      
      m.src = p.mask; // point of discovery ;)
      if (p[1])
        m.dst = p[1];
      
      if (m.dst && m.src)
        log (L_DISC, "SERVER MASKS DISCOVERED: %s (%02x) -> %s (%02x)\n",
                      p.dst_msn, m.src, p.src_msn, m.dst);
      
      // other side needs information? need to ack..
      if (!(p[1] && p[2]))
        {
          log (L_DISC, "ack'ing discovery packet (%02x, %02x)\n", m.src, m.dst);
          once.set (p.dst_msn, p.src_msn, m.dst ? m.dst : IEM_ALL);
          once.put (MSG_DISCOVERY);
          once.put (m.src);
          once.put (m.dst);
          while (once.size () < MAX_PACKET)
            once.put (0);
        }
      
      tx_now ();
    }
  else if (p[0] == MSG_CFG)
    switch (p[1] & ~CFG_REPLY)
      {
        case CFG_POLL:
          log (L_WARN, "SERVER POLL %s (%s) protocolversion %d, %d connections\n",
                        p.src_msn, &p[4], p[2], p[3]);
          if (!(p[1] & CFG_REPLY))
            send_ctrl_poll (p.dst_msn, p.src_msn, CFG_REPLY);
          break;
        
        case CFG_MSG:
          {
            char dt[40];
            time_t tt = time (0);
            
            strftime (dt, 40, "%a %H:%M:%S", localtime (&tt));
            printf ("\r%s MESSAGE FROM %s: '", dt, p.src_msn);
            packet::const_iterator i = p.begin (); i++; i++;
            while (i != p.end ())
              putc (*i++, stdout);
            printf ("'\n");
            break;
          }
        
        default:
          log (L_WARN, "WARNING: unknown MSG_CFG (%02x) received!!\n", p[1]);
      }
  else if (p[0] == MSG_RST)
    {
      log (L_WARN, "warning, RESET received from %s\n", p.src_msn);
      line_control.command_all (&connection::reset, p.src_msn);
    }
  else if ((p[0] & MSG_MASK) == MSG_SYN)
    {
      log (L_INFO, "connection request ignored from %s, pver %d, port %d -> %d\n",
           p.src_msn, p[1], p[2], p[3]);
      send_rst (p);
    }
  else
    {
      log (L_WARN, "WARNING: unknown config packet %02x received!!\n", (int)p[0]);
      send_rst (p);
    }
  
  lock.unlock ();
}

void mask_map::dump ()
{
  const_iterator i = begin ();
  
  while (i != end ())
    {
      printf ("%s: %02x -> %02x\n", i->first.c_str (), i->second.src, i->second.dst);
      ++i;
    }
}

ie_mask mask_map::dst_mask (const char *dst)
{
  ie_masks &m = operator [](dst);
  
  if (m.src && m.dst)
    return m.dst;
  
  static timestamp last_discover;
  
  if (last_discover.past ())
    {
      packet &p = control_connection.once;
      
      control_connection.lock.lock ();
      last_discover.stamp (control_connection.disc_timeout ());
      control_connection.disc_timeout.next ();
      p.set (cfg_msn, dst, IEM_ALL);
      p.put (MSG_DISCOVERY);
      p.put (m.src);
      p.put (m.dst);
      while (p.size () < MAX_PACKET)
        p.put (0);
      
      control_connection.once = p;
      control_connection.tx_now ();
      control_connection.lock.unlock ();
    }
  
  return IEM_NONE;
}

#define YYERROR_VERBOSE

void yyerror (const char *msg)
{
  fprintf (stderr, "%s, line %d: %s\n", yyfilename, yylineno, msg);
}

// connection setup parameters
static plug *plg;
static char id[40];
static msn src;
static u8 src_port;
static msn dst;
static u8 dst_port;

%}

%union {
   char str[1024];
}

%token <str> NUMERAL STRING

%token MODEMDEV CTRLDEV CFG_MSN SERVER_IDENT ALIAS
%token CONNECTION SRC DST MSN PORT TTY_PLUG
%token CONNECT LISTEN STOP QUIT
%token INFO LIST DUMP LOGOPT
%token POLL MSG
%token EOL

%type <str> STRING SRC DST

%start config_file

%%

config_file:	decls;

eol:		/* empty */ | EOL eol;

decls:		/* empty */ | decls decl;

decl:		  MODEMDEV STRING EOL
			{ line_control.add_modemdev ($2); }
		| CTRLDEV STRING EOL
			{ line_control.set_ctrldev ($2); }
		| CFG_MSN NUMERAL NUMERAL NUMERAL EOL
			{ strcpy (cfg_country, $2); strcpy (cfg_area, $3); strcpy (cfg_msn, $4);
			  normalize_msn (cfg_msn); }
		| SERVER_IDENT STRING EOL
			{ strcpy (servername, $2); }
		| connection
		| command
		| EOL
		;

connection:	CONNECTION STRING eol
                	{ src_port = dst_port = 0; plg = 0; src[0] = 0; dst[0] = 0; strcpy (id, $2); }
                '{' eol connection_tags '}' EOL
                	{
			  if (!plg)
                	    yyerror ("plug tag missing");
                	  else if (!src[0])
                	    yyerror ("src tag missing");
                	  else if (!dst[0])
                	    yyerror ("dst tag missing");
                	  else
                	    {
                	      if (!src_port)
                	        {
                	          log (L_WARN, "WARNING: defaulting src_port to %d\n", PORT_ITCP);
                	          src_port = PORT_ITCP;
                	        }
                	      if (!dst_port)
                	        {
                	          log (L_WARN, "WARNING: defaulting dst_port to %d\n", PORT_ITCP);
                	          dst_port = PORT_ITCP;
                	        }
                	      normalize_msn (src);
                	      normalize_msn (dst);
                	      connection *nc = new itcp_connection (src, src_port, dst, dst_port, *plg);
                	      printf ("adding itcp_connection %s: ", id); nc->dump ();
			      connections[ strdup(id) ] = nc;
			    }
                	}
                ;

connection_tags:  /* empty */ | connection_tag eol connection_tags;

connection_tag:	  plug
		| SRC MSN NUMERAL
			{ strcpy (src, $3); }
		| DST MSN NUMERAL
			{ strcpy (dst, $3); }
		| SRC PORT NUMERAL
			{ src_port = atoi ($3); }
		| DST PORT NUMERAL
			{ dst_port = atoi ($3); }
		;

plug:		TTY_PLUG STRING
			{ if (plg) delete plg; plg = new tty_plug ($2); }
		;

command:	  CONNECT STRING EOL
			{ con_apply ($2, &connection::connect, "connect"); }
		| LISTEN STRING EOL
			{ con_apply ($2, &connection::listen, "listen"); }
		| STOP STRING EOL
			{ con_apply ($2, &connection::stop, "stop"); }
		| INFO STRING EOL
			{ con_apply ($2, &connection::dump); }
		| LIST EOL
			{
			  con_map::iterator p = connections.begin ();
			  while (p != connections.end ())
			    {
			      printf ("%s: ", p->first.c_str ());
			      p->second->dump ();
			      ++p;
			    }
			}
		| POLL NUMERAL EOL
			{ normalize_msn ($2); control_connection.cfg_poll ($2); };
		| DUMP EOL
			{ mask_map.dump (); }
		| MSG NUMERAL STRING
			{ strcpy (last_chat, $2); normalize_msn (last_chat);
			  control_connection.chat (last_chat, $3); }
		| MSG STRING
			{ control_connection.chat (last_chat, $2); }
		| LOGOPT EOL
			{ parse_logopt (0); }
		| LOGOPT logopts
		| QUIT EOL
			{ exit (0); }
		;

logopts:	  STRING logopts
			{ parse_logopt ($1); }
		| EOL
		;

%%

int main (int argc, char *argv[])
{
  assert (argc == 2);
  
  init_debug (argv);
  
#if YYDEBUG != 0
  yydebug = 1;
#endif
  
  yyin = fopen (argv [1], "r");
  assert (yyin);
  yyfilename = argv [1];
  yylineno = 1;
  
  line_control.add_connection (&control_connection);
  
  for(;;)
    yyparse ();
  
  return 0;
}

